#!/usr/bin/perl -w
#
# chronoGrapher.pl v1.3.3
# Last modified June 17, 2005 16:07
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#	1) A nexus-formatted treefile
#	2) A tab-delimited text file with user-input nodal divergence times (node number<tab>date)
#
# Output:
#   1) A nexus-formatted treefile with branch lengths calculated to fit nodal divergence times
#	2) (Optional) )A tab-delimited text file giving interpolated and/or corrected nodal divergence times
#
# Usage: chronoGrapher.pl [-d<filename>] [-f] [-p] [-t<filename>] [-z<number>] [-v] [-h]
#   options: -d<filename> = tab-delimited textfile with nodal divergence times (node number<tab>date)
#                           (default = datefile.txt)
#	         -f = fix any negative or zero length branches
#	         -i = do not interpolate divergence dates for nodes without date estimates
#			 -p = protect date of root node when correcting negative or zero length branches
#   		 -t<filename> = treefile containing trees in NEXUS format
#                           (default = treefile.tre)
#   		 -z<number> = correction value for zero branch lengths, whether user-input or from
#						  fixing process (default = 0.1); only operative with -f flag
#            -v = verbose output
#            -h = print this message and quit

use strict;

# Set defaults
	# Input variables
		my $treefile = "treefile.tre";
		my $datefile = "datefile.txt";
		my $block = 1;
		my $treeBlock = 0;
		my $interpolate = 1;
	
	# Tree variables
		my $SupportFlag;
		my %branchLength;
		my $maxClades;

	# Date variables
		my (%divergeDate, %interpolDate, %fixedDate);
		my (@nodeList, @nodeErrors, @dateErrors, @missingNodes, %missingDate, %datePresent);
		my $maxNode = 0;
		my $negFix = 0;
		my $protectRoot = 0;
		my $topologyFix = 0.1;

	# Output variables
		my @chronoTrees;

	# Miscellaneous variables
		my $verbose = 0;
		my $debug = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-d([\w+\.?\w+?]*)/)
		{
		$datefile = $1;
		}
	elsif ($ARGV[$i] =~ /-f/)
		{
		$negFix = 1;
		}
	elsif ($ARGV[$i] =~ /-i/)
		{
		$interpolate = 0;
		}
	elsif ($ARGV[$i] =~ /-p/)
		{
		$protectRoot = 1;
		}
	elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
		{
		$treefile = $1;
		}
	elsif ($ARGV[$i] =~ /-z([\d+\.?\d+?]*)/)
		{
		$topologyFix = $1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] =~ /-x/)
		{
		$debug = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: chronoGrapher.pl [-d<filename>] [-f] [-p] [-t<filename>] [-z<number>] [-v] [-h]\n";
		print "Options: -d<filename> = tab-delimited textfile with nodal divergence times (node number<tab>date)\n";
		print "                        (default = datefile.txt)\n";
		print "         -f = fix any negative or zero length branches\n";
		print "         -i = do not interpolate divergence dates for nodes without date estimates\n";
		print "         -p = protect date of root node when correcting negative or zero length branches\n";
		print "         -t<filename> = file containing trees in NEXUS format\n";
		print "                        (default = treefile.tre)\n";
		print "         -z<number> = correction value for zero branch lengths, whether user-input or from\n";
		print "                      fixing process (default = 0.1); only operative with -f flag\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "chronoGrapher.pl [-d<filename>] [-f] [-p] [-t<filename>] [-z<number>] [-v] [-h]\n";
		exit(1); 
		}
	}

(my $outTreeFile = $treefile) =~ s/\.\w*$//;
	$outTreeFile .= "_chrono.tre";
(my $outDateFile = $datefile) =~ s/\.\w*$//;
	$outDateFile .= "_chrono.txt";

print "Reading in nodal divergence times ...\n";
	setLineBreak($datefile);
	open (DATE, "<$datefile") or die "Cannot open textfile containing nodal divergence times, $datefile\n";
		while (<DATE>)
			{
			chomp;
			next unless ($_);
			print "\t$_\n" if ($debug);
			my ($node, $date) = split(/\t/);
				(my $nodeNum = $node) =~ s/\D//g;
			# Check node and date for errors
				print "\t\tNode: " if ($debug);			
					if ($nodeNum eq "" or not defined $nodeNum)
						{
						push @nodeErrors, $node;
						print "Error with node formatting\n" if ($debug);
						}
					else
						{
						push @nodeList, $nodeNum;
						$maxNode = $nodeNum if ($nodeNum > $maxNode);
						print "$node ($nodeNum)\n" if ($debug);
						}
				print "\t\tDate: " if ($debug);			
					if (defined $date and $date =~ m/\d+\.?\d?/)
						{
						$divergeDate{$nodeNum} = $date;
						push @nodeList, $nodeNum;
						$datePresent{$nodeNum} = 1;
						print "$date\n" if ($debug);
						}
					else
						{
						push @missingNodes, "\"$node\"" unless ($missingDate{$nodeNum});
						$missingDate{$nodeNum} = 1;
						print "missing\n" if ($debug);
						}
			}

	# Check for errors
		if (@nodeErrors)
			{
			print "\n\tERROR: Formatting errors for the following node names:\n";
			print "\t\t$_\n" foreach (@nodeErrors);
			exit(1);
			}
	
	# Check for missing node numbers
		@nodeList = sort {$a <=> $b} @nodeList;
		for (my $i = 1; $i < $maxNode; $i++)
			{
			push @missingNodes, "Node $i" unless ($datePresent{$i} or $missingDate{$i});
			}
		if (@missingNodes)
			{
			print "\n\tThe following ".scalar(@missingNodes)." nodes (of $maxNode total) are missing date estimates:\n";
			print "\t\t$_\n" foreach (@missingNodes);
			if ($interpolate)
				{
				print "\n\t\tDates will be interpolated based on other dates present\n";
				}
			else
				{
				print "\n\tERROR: Interpolate of missing dates not requested; cannot continue\n";
				exit(1);
				}
			}

print "\nReading in tree descriptions ...\n";
	my $readZero = time;
	setLineBreak($treefile);
	open (fTREES, "<$treefile") or die "Cannot open treefile $treefile\n";		# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
		my (@AllTrees, %treeNames, $treeDescrip, $treeFlag, $treeCount);
		my (@taxonList, %taxonPresent);
		my $NumChar = 0;
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<fTREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$treeDescrip .= $_;
				if ($_ =~ m/:/)
					{
					$SupportFlag = 1;
					}
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		#Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						push @AllTrees, $tree;
						print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close fTREES;

	my $readTime = time - $readZero;
	print "\tTime taken: $readTime second(s)\n";
	die "\nERROR: No trees read in from file $treefile\n" if (not @AllTrees);

# Calculate branch lengths for chronograph, interpolating and correcting dates as necessary
	print "\nCalculating branch lengths for chronograph ...\n";
	$treeCount = 0;
	foreach my $tree (@AllTrees)
		{
		my $treeZero = time;
		$treeCount++;
		my $negFlag = 0;
#		my $newST;
		print "\tProcessing tree #$treeCount: $treeNames{$treeCount}\n";
		my $cladeCount = ($tree =~ tr/(//);
			if ($cladeCount != $maxNode)
				{
				print "\t\tERROR: Number of nodes on tree ($cladeCount) does not match number of divergence times ($maxNode); skipping to next tree\n";
				next;
				}
		$tree =~ s/:\d+\.?\d*e?\-?\d*//g;	# Remove any existing branch length information

		# Do initial parse of tree to get structure and check for any negative or zero branch lengths
			print "\t\tInitially parsing tree ...\n";
				my ($position, $CurrentClade) = (0, 0);
				my (%ParentClade, %open, %daughterClade, %cladeSize, %FCbalance, %numDaughters);
				my $CladeID = 0;
				my $OldCurrent;
				
#				my $numNegBL = 0;
			
			while ($position < length($tree))	# Parse tree from root to tips
				{
				my $offset = 1;
				if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$CladeID++;
					print "\t\t\tClade #$CladeID\n" if ($debug);
					$open{$CladeID} = 1;
					$ParentClade{$CladeID} = $CurrentClade;
					push @ {$daughterClade{$CurrentClade} }, $CladeID;	# Assign new clade as daughter clade to parent clade
						$numDaughters{$CurrentClade}++;
					$CurrentClade = $CladeID;
#					unless ($CladeID == 1)
#						{
#						$branchLength{$CladeID} = sprintf ("%.1f", $divergeDate{$ParentClade{$CladeID}} - $divergeDate{$CladeID});
#						if ($debug)
#							{
#							print "\t\t\tParental divergence time: $divergeDate{$ParentClade{$CladeID}}\n";
#							print "\t\t\tDivergence time: $divergeDate{$CladeID}\n";
#							print "\t\t\tBranch length: $branchLength{$CladeID}\n";
#							}
#						if ($branchLength{$CladeID} <= 0)
#							{
#							$numNegBL++;
#							print "\t\t\tERROR: Negative branch length for node $CurrentClade\n";
#							}
#						}
					}
				elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};

					print "\t\t\tFinishing clade #$OldCurrent\n\t\t\t\tSize: $cladeSize{$OldCurrent}\n" if ($debug);

					# Calculate Fusco and Cronk I balance metric (corrected for polytomies)
#						$FCbalance{$OldCurrent} = 0.5;	# Default value for all clades
#						if ($cladeSize{$OldCurrent} > 3)	# Recalculate value for clades > 3, where metric is relevant
#							{
#							my $littleM = $cladeSize{$OldCurrent} / $numDaughters{$OldCurrent};	# Minimum size of largest branch
#								$littleM = int($littleM) + 1 unless ($littleM == int($littleM));	# Get smallest integer >= S / d
#							my $bigM = $cladeSize{$OldCurrent} - $numDaughters{$OldCurrent} + 1;	# Maximum size for biggest branch
#							my $bigB = 0;	# Actual size of biggest branch
#								foreach my $daughter (@ {$daughterClade{$OldCurrent} })
#									{
#									$bigB = $cladeSize{$daughter} if ($cladeSize{$daughter} > $bigB);
#									}
#							
#							print "\t\t\t\tm: $littleM\tM: $bigM\tB: $bigB\td: $numDaughters{$OldCurrent}\n" if ($debug);
#							if ($bigM == $littleM)	# Traps against division by zero error in equation; node is as balanced as could be expected
#								{
#								$FCbalance{$OldCurrent} = 0; 
#								}
#							else
#								{
#								$FCbalance{$OldCurrent} = ($bigB - $littleM) / ($bigM - $littleM); 
#								}
#							}
#						print "\t\t\t\tI: $FCbalance{$OldCurrent}\n" if ($debug);
					}
				elsif (substr($tree, $position, 1) eq "[")	# Skip over embedded comment
					{
					until (substr($tree, $position+$offset, 1) eq "]" or $position+$offset >= length($tree))
						{
						$offset++;
						}
					$position += $offset-1;
					}
				elsif (substr($tree, $position, 1) eq ",")
					{
					# Do nothing
					}
				else	# Determine membership of clade
					{
					my $prevChar = substr($tree,$position-1, 1);
					until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or ($position+$offset) >= length($tree))
						{
						$offset++;
						}
					unless ($prevChar eq ")")	# Safety to block any node labels, which would otherwise appear to be a taxon
						{
						my $taxon = substr($tree, $position, $offset);
						$numDaughters{$CurrentClade}++;
#							my $brLength = sprintf ("%.1f", $divergeDate{$CurrentClade});
							if ($debug)
								{
								print "\t\t\t\tTaxon: $taxon\n";
#								print "\t\t\tBranch length: $brLength\n";
								}
						for (1..$CurrentClade)	# Increment size counter for all open clades
							{
							$cladeSize{$_}++ if ($open{$_});
							}
						}
					$position += $offset-1;
					}
				$position++;
				}

			$maxClades = $CladeID;
			
			# Interpolate any missing dates, if necessary
				# Missing dates are interpolated according only to the log-log formula found in Purivs (1995)
					# Although empirical results show that log-log is only accurate for balanced trees, squared-squared variant of formula (which is better for unblanaced trees and is coded (as weighted average with log-log date based on Fusco-Cronk balance) is not used because it will given interpolated dates of 0 when the parent clade is significantly larger than the target clade
				# Interpolation proceeds first by using a weighted average of up to five ancestral nodes (inversely weighted according to distance: 5, 4, 3, 2, and 1) plus immediate daughter nodes (weight = 5)
				# If no suitable nodes are found, first available ancestral node is used
				
				if (@missingNodes)
					{
					print "\n\t\tInterpolating missing dates ...\n";
					for (my $treeNode = 1; $treeNode <= $maxClades; $treeNode++)
						{
						next if ($divergeDate{$treeNode});
						print "\t\t\tNode: $treeNode (size: $cladeSize{$treeNode})\n" if ($debug);
						# Interpolate date based on ancestral nodes
							print "\t\t\t\tInterpolating from ancestral nodes ...\n" if ($debug);
							my $ancClade = $ParentClade{$treeNode};
							my $weightedN = 0;
							for (my $ancLevel = 5; $ancLevel >= 1; $ancLevel--)
								{
								printf "\t\t\t\t\tAncestral level %s (node: $ancClade; size: $cladeSize{$ancClade}", 6-$ancLevel if ($debug);
								if ($divergeDate{$ancClade})
									{
									print "; age: $divergeDate{$ancClade})\n" if ($debug);
									$weightedN += $ancLevel;
									
									# Calculate interpolated dates using Purvis log-log formula
										my $logDate = $divergeDate{$ancClade} * (log($cladeSize{$treeNode}) / log(10)) / (log($cladeSize{$ancClade}) / log(10));
#											printf "\t\t\t\t\t\tLog date %.1f * %.3f\n", $logDate, 1 - $FCbalance{$ancClade} if ($debug);
#											$logDate *= (1 - $FCbalance{$ancClade});
#										my $squareDate = $divergeDate{$ancClade} * ($cladeSize{$treeNode})**2 / ($cladeSize{$ancClade})**2;
#											printf "\t\t\t\t\t\tSquared date %.1f * %.3f\n", $squareDate, $FCbalance{$ancClade} if ($debug);
#											$squareDate *= $FCbalance{$ancClade};									
										$interpolDate{$treeNode} += $logDate * $ancLevel;
#											$interpolDate{$treeNode} += ($logDate + $squaredDate) * $ancLevel;
										printf "\t\t\t\t\t\tInterpolated date %.1f * %s\n", $logDate, $ancLevel if ($debug);
#											$interpolDate{$treeNode} += ($logDate + $squareDate) * $ancLevel;
#											printf "\t\t\t\t\t\tInterpolated date %.1f * %s\n", $logDate + $squareDate, $ancLevel if ($debug);
									}
								else
									{
									print "\n" if ($debug);
									}
								last if ($ancClade == 1);	# Stop looping if already at root
								$ancClade = $ParentClade{$ancClade};								
								}
							print "\t\t\t\t\tNo suitable ancestral nodes\n" if ($debug and not $weightedN);

						# Interpolate date based on descendent nodes
							print "\n\t\t\t\tInterpolating from daughter nodes ...\n" if ($debug);
							if (defined @ {$daughterClade{$treeNode} })
								{
								my $daughterDate;
								my $numDaughterDate = 0;
								foreach my $daughter (@ {$daughterClade{$treeNode} })
									{
									next unless ($divergeDate{$daughter});
									$numDaughterDate++;
									print "\t\t\t\t\tDaughter node $numDaughterDate (node: $daughter; size: $cladeSize{$daughter}; age: $divergeDate{$daughter})\n" if ($debug);
									
									my $logDate = $divergeDate{$daughter} * (log($cladeSize{$treeNode}) / log(10)) / (log($cladeSize{$daughter}) / log(10));
#										printf "\t\t\t\t\t\tLog date %.1f * %.3f\n", $logDate, 1 - $FCbalance{$ancClade} if ($debug);
#										$logDate *= (1 - $FCbalance{$treeNode});
#									my $squareDate = $divergeDate{$daughter} * ($cladeSize{$treeNode})**2 / ($cladeSize{$daughter})**2;
#										printf "\t\t\t\t\t\tSquared date %.1f * %.3f\n", $squareDate, $FCbalance{$ancClade} if ($debug);
#										$squareDate *= $FCbalance{$treeNode};									
									$daughterDate += $logDate;
#										$daughterDate += ($logDate + $squareDate);
									printf "\t\t\t\t\t\tInterpolated date %.1f\n", $logDate if ($debug);
#										printf "\t\t\t\t\t\tInterpolated date %.1f\n", $logDate + $squareDate if ($debug);
									}
								if ($numDaughterDate)
									{
									$weightedN += 5;
									$interpolDate{$treeNode} += ($daughterDate / $numDaughterDate) * 5;
									printf "\n\t\t\t\t\tAverage daughter interpolated date %.1f * 5\n", $daughterDate / $numDaughterDate if ($debug);
									}
								else
									{
									print "\t\t\t\t\tNo dated daughter nodes available\n" if ($debug);
									}
								}
							else
								{
								print "\t\t\t\t\tNo daughter nodes available\n" if ($debug);
								}

						# Calculate weighted average date
							if ($weightedN)
								{
								$interpolDate{$treeNode} = sprintf("%.1f", $interpolDate{$treeNode} / $weightedN);
								}
							else	# Interpolate date based on first available ancestral node
								{
								print "\n\t\t\t\tInterpolating from first available ancestral node ...\n" if ($debug);
								until ($divergeDate{$ancClade} or $ancClade == 1)
									{
									print "\t\t\t\t\tChecking node $ancClade\n" if ($debug);
									$ancClade = $ParentClade{$ancClade};
									}
								
								my $logDate = $divergeDate{$ancClade} * (log($cladeSize{$treeNode}) / log(10)) / (log($cladeSize{$ancClade}) / log(10));
#									printf "\t\t\t\t\t\tLog date %.1f * %.3f\n", $logDate, 1 - $FCbalance{$ancClade} if ($debug);
#									$logDate *= (1 - $FCbalance{$ancClade});
#								my $squareDate = $divergeDate{$ancClade} * ($cladeSize{$treeNode})**2 / ($cladeSize{$ancClade})**2;
#									printf "\t\t\t\t\t\tSquared date %.1f * %.3f\n", $squareDate, $FCbalance{$ancClade} if ($debug);
#									$squareDate *= $FCbalance{$ancClade};									
								$interpolDate{$treeNode} = sprintf("%.1f", $logDate);
#									$interpolDate{$treeNode} = sprintf("%.1f", $logDate + $squareDate);
								print "\n\t\t\t\t\tInterpolated date: $interpolDate{$treeNode}\n" if ($debug);
								}
								
							if ($interpolDate{$treeNode} < 0)
								{
								print "\t\t\t\t\tDate less than 0: setting to zero\n" if ($debug);
								$interpolDate{$treeNode} = 0;
								}
								
							print "\n\t\t\t\tFinal interpolated date: $interpolDate{$treeNode} ($weightedN)\n\n" if ($debug);
						}
					}

			# Save orginial and interpolated dates to new hash
				for (my $treeNode = 1; $treeNode <= $maxClades; $treeNode++)
					{
					if ($divergeDate{$treeNode})
						{
						$fixedDate{$treeNode} = sprintf("%.1f", $divergeDate{$treeNode});
						}
					else
						{
						$fixedDate{$treeNode} = sprintf("%.1f", $interpolDate{$treeNode});
						}
					}

			# Fix any negative or zero length branches, if desired / necessary
				# Procedure goes through nodes of tree sequentially
					# For branch lengths <= 0, new dates are assigned based on either:
						# 1) average date of node and all older daughter nodes, with daughter nodes contributing independently to average; nodes deviate from average by a correction measure to retain topology
						# 2) root node maintains date if protected, with all older daughter nodes receiving the date of the root node minus the correction measure
				# Procedure loops until no more branch lengths <= 0 are found

				print "\n\t\tChecking for negative or zero length branches ...\n";
					print "\t\t\tWill not be fixed and tree will not be saved\n" if ($negFix == 0);
					print "\n" if ($negFix == 0);
			
				my $numNegBL = 1;
				my $iLoop = 0;

				until ($numNegBL == 0 or $negFlag == 1)	# Loop until negative and zero length branches are fixed
					{
					$iLoop++;
					print "\t\t\tLoop: $iLoop\n";
					$numNegBL = 0;
					for (my $parentNode = 1; $parentNode <= $maxClades; $parentNode++)
						{
						my $sumDaughterAges = 0;
						my $numNegDaughter = 0;
						my @negDaughters;

						print "\t\t\t\tClade: $parentNode\n" if ($debug);
						print "\t\t\t\t\tDivergence time: $fixedDate{$parentNode}\n" if ($debug);
						
						foreach my $daughterNode ( @ {$daughterClade{$parentNode} })	# Check branches leading to each daughter
							{
							print "\t\t\t\t\tDaughter ($daughterNode) divergence time: $fixedDate{$daughterNode}\n" if ($debug);
							if ($fixedDate{$daughterNode} >= $fixedDate{$parentNode})
								{
								print "\t\t\t\t\t\tNegative or zero length branch\n" if ($debug);
								$sumDaughterAges += $fixedDate{$daughterNode};
								$numNegDaughter++;
								push @negDaughters, $daughterNode;
								if ($negFix == 0)
									{
									print "\t\t\t\tERROR: Negative or zero length branch for node $daughterNode\n" if (not $debug);
									$negFlag = 1;
									}
								}
							if ($fixedDate{$daughterNode} <= 0)
								{
								print "\t\t\t\t\t\tDate less than or equal to 0: setting to $topologyFix\n" if ($debug);
								$fixedDate{$daughterNode} = $topologyFix;
								}
							}
						if ($numNegDaughter and $negFix)	# At least one branch length is <= 0; fix by taking average age of all nodes contributing to negative branch length(s); all nodes contribute equally to average so no accounting for phylogenetic structure is made
							{
							$numNegBL += scalar(@negDaughters);
							if ($parentNode == 1 and $protectRoot)
								{
								print "\n\t\t\t\t\tNew divergence time: $fixedDate{$parentNode}\n" if ($debug);
								foreach my $daughter (@negDaughters)
									{
									$fixedDate{$daughter} = sprintf ("%.1f", $fixedDate{$parentNode} - $topologyFix);
									print "\t\t\t\t\tNew daughter ($daughter) divergence time: $fixedDate{$daughter}\n" if ($debug);
									if ($fixedDate{$daughter} < 0)
										{
										print "\t\t\t\t\t\tDate less than 0: setting to zero\n" if ($debug);
										$fixedDate{$daughter} = 0;
										}
									}
								}
							else
								{
								my $avgAge = ($fixedDate{$parentNode} + $sumDaughterAges) / ($numNegDaughter + 1);
								$fixedDate{$parentNode} = sprintf ("%.1f", $avgAge + $topologyFix);
									print "\n\t\t\t\t\tNew divergence time: $fixedDate{$parentNode}\n" if ($debug);
								foreach my $daughter (@negDaughters)
									{
									$fixedDate{$daughter} = sprintf ("%.1f", $avgAge - $topologyFix);
									print "\t\t\t\t\tNew daughter ($daughter) divergence time: $fixedDate{$daughter}\n" if ($debug);
									if ($fixedDate{$daughter} <= 0)
										{
										print "\t\t\t\t\t\tDate less than or equal to 0: setting to $topologyFix\n" if ($debug);
										$fixedDate{$daughter} = $topologyFix;
										}
									}
								}
							}
						}
					print "\n\t\t\t\t$numNegBL negative or zero length branch(es) detected\n\n" if ($debug);
					}
				
				# Reparse tree to yield chronogram
					if ($negFlag)
						{
						print "\n\t\tERROR: Tree will not be saved to file\n";
						}
					else
						{
						print "\n\t\tWriting corrected tree to file ...\n";

						# Reset tree parsing variables
							my ($position, $CurrentClade) = (0, 0);
							my (%ParentClade, %open);
							my $CladeID = 0;
							my $OldCurrent;
			
						my $newST;
							
						while ($position < length($tree))	# Parse tree from root to tips
							{
							my $offset = 1;
							if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
								{
								$CladeID++;
								print "\t\t\tClade #$CladeID\n" if ($debug);
								$ParentClade{$CladeID} = $CurrentClade;
								$CurrentClade = $CladeID;
								$newST .= "(";
								unless ($CladeID == 1)
									{
									$branchLength{$CladeID} = sprintf ("%.1f", $fixedDate{$ParentClade{$CladeID}} - $fixedDate{$CladeID});
									if ($debug)
										{
										print "\t\t\t\tParental divergence time: $fixedDate{$ParentClade{$CladeID}}\n";
										print "\t\t\t\tDivergence time: $fixedDate{$CladeID}\n";
										print "\t\t\t\tBranch length: $branchLength{$CladeID}\n";
										}
									}
								print "\t\t\t\tDivergence time: $fixedDate{$CladeID}\n" if ($debug and $CladeID == 1);
								}
							elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
								{
								$open{$CurrentClade} = 0;
								$OldCurrent = $CurrentClade;
								$CurrentClade = $ParentClade{$OldCurrent};
								$newST .= ")";
								$newST .= ":$branchLength{$OldCurrent}" unless ($OldCurrent == 1);
								}
							elsif (substr($tree, $position, 1) eq "[")	# Skip over embedded comment
								{
								until (substr($tree, $position+$offset, 1) eq "]" or $position+$offset >= length($tree))
									{
									$offset++;
									}
								$position += $offset-1;
								}
							elsif (substr($tree, $position, 1) eq ",")
								{
								$newST .= substr($tree, $position, 1);
								}
							else	# Determine membership of clade
								{
								my $prevChar = substr($tree,$position-1, 1);
								until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or ($position+$offset) >= length($tree))
									{
									$offset++;
									}
								unless ($prevChar eq ")")	# Safety to block any node labels, which would otherwise appear to be a taxon
									{
									my $taxon = substr($tree, $position, $offset);
										my $brLength = sprintf ("%.1f", $fixedDate{$CurrentClade});
									if ($debug)
										{
										print "\t\t\tTaxon: $taxon\n";
										print "\t\t\t\tBranch length: $brLength\n";
										}
									$newST .= "$taxon:$brLength";
									}
								$position += $offset-1;
								}
							$position++;
							}
						push @chronoTrees, $newST;
						}

		my $treeTime = time - $treeZero;
		print "\n\t\tTree processed in $treeTime seconds\n";
		}

# Print updated datefile if necessary
	if (@missingNodes or $negFix)
		{
		open (DATE, ">$outDateFile") or die "Cannot open output file $outDateFile to write nodal divergence dates.\n";
			print DATE "Node\tInput date";
				print DATE "\tInterpolated date" if (@missingNodes);
				print DATE "\tCorrected date" if ($negFix);
				print DATE "\n";
			
			for (1..$maxClades)
				{
				print DATE "$_\t";
				print DATE "$divergeDate{$_}" if ($divergeDate{$_});
					print DATE "\t";
				print DATE "$interpolDate{$_}" if (@missingNodes and $interpolDate{$_});
					print DATE "\t" if (@missingNodes);
				print DATE "$fixedDate{$_}" if ($negFix);
					print DATE "\n";
				}
		close DATE;
		}

# Print (any) chronographs
	if (@chronoTrees)
		{
		my $treeCount = 0;
		open (TREE, ">$outTreeFile") or die "Cannot open output file $outTreeFile to write tree description(s).\n";
			print TREE "#NEXUS\n\nbegin trees;\n";
			foreach my $cTree (@chronoTrees)
				{
				$treeCount++;
				print TREE "\tTREE $treeNames{$treeCount} = $cTree;\n";
				}
			print TREE "end;\n";
		close TREE;
		}

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
# 	v1.3.3 (June 17, 2005)
#		- minimum age of nodes during date fixing process now equal to correction value for zero or negative
#		  branch lengths (-z); prevents terminal branches from having a zero branch length
#			- NOTE: interpolated dates for nodes still have lower bound of zero; new value only applies during
#					date fixing process
#		- minor, cosmetic bug fix relating to output file
#
# 	v1.3.2 (February 28, 2005)
#		- enforced minimum age of zero throughout to prevent terminal ages from being negative
#
# 	v1.3.1 (February 7, 2005)
#		- minor bug fixes
#
# 	v1.3 (July 28, 2004)
#		- added ability to interpolate missing divergence times using log-log formula
#	 	  from Purvis (1995)
#
# 	v1.2 (June 30, 2004)
#		- added ability to correct for negative or zero branch lengths
#
# 	v1.1 (June 14, 2004)
#		- minor bug fixes
#
# 	v1.0 (April 27, 2004)
#		- initial release

